/*-------------------------------------------------------------
  module: poly_ntl3.0c.C
  author: Angel L. Diaz, E. Kaltofen
  date: 8/14/96, Sep 4, 1998

  Test the black box polynomial classes and its sub components.
-------------------------------------------------------------*/

// arithmetic includes
#include "PlugIns/ntl3.0c.h"
#include "BlackBox/CommonObjects/bbpoly.h"

NTLzz_p  p1(NTLzz_p *); 
NTLzz_pX p2(NTLzz_pX *);

int main() {
   MP_INT MPCard;
   mpz_init_set_ui( &MPCard,  65521 );
   NTLInitEnv(&MPCard );

   NTLzz_pX Poly1( (long)21, 2 ), Poly2( (long)40, 1 ), Poly3;
   Poly3 = Poly1 + Poly2 + (long) 11;
   NTLzz_p Elm( (long) 100 );
   cout << "poly  : Poly3 = " << Poly3 << endl;

   unsigned int NumVars = 1;
   unsigned int Deg = 2;
   unsigned int Prob = 1;
   BlackBoxPoly< NTLzz_p,NTLzz_pX> BBPoly1( Poly3, NumVars, Deg, Prob );
   cout << "poly  : BBPoly1.Deg() = " << BBPoly1.Deg() << endl;
   cout << "poly  : BBPoly1.NumVars() = " << BBPoly1.NumVars() << endl;
   cout << "poly  : BBPoly1.Prob() = " << BBPoly1.Prob() << endl;
   cout << "poly  : BBPoly1( &Elm ) = " << *BBPoly1( &Elm ) << endl;

   BlackBoxUPoly< NTLzz_p > BBUPoly1( p1, NumVars, Deg, Prob );
   cout << "poly  : BBUPoly1( &Elm ) = " << *BBUPoly1( &Elm ) << endl;

   BlackBoxUPolyEx< NTLzz_p,NTLzz_pX > BBUPoly2( p1, p2, NumVars, Deg, Prob );
   cout << "poly  : BBUPoly2( &Elm ) = " << *BBUPoly2( &Elm ) << endl;
   NTLzz_pX Elm2( (long)1, 1 );
   cout << "poly  : BBUPoly2( &Elm2 ) = " << *BBUPoly2( &Elm2 ) << endl;
 
   NTLCleanUpEnv();
}

// p1 := 21*x^2 + 40*x + 1 
NTLzz_p p1(
   NTLzz_p *pnts
   ) 
{
   NTLzz_p X = pnts[0];
   return( X * X * (long) 21 + X * (long) 40 + (long) 1 );
}

NTLzz_pX p2(
   NTLzz_pX *pnts
   )
{
   NTLzz_pX X = pnts[0];
   return( X * X * (long) 21 + X * (long) 40 + (long) 1 );
}

