#ifndef ntl_matrix__H
#define ntl_matrix__H

#include "tools.h"
#include "ntl_vector.h"


// matrix templates


#define ntl_matrix_decl(T,vec_T,vec_vec_T,mat_T)  \
class mat_T {  \
private:  \
  \
   vec_vec_T rep;  \
   long numcols;  \
  \
public:  \
  \
   mat_T() { numcols = 0; }  \
   mat_T(const mat_T& a);  \
   mat_T& operator=(const mat_T& a);  \
   ~mat_T() { }  \
  \
   mat_T(INIT_SIZE_TYPE, long n, long m);  \
  \
   void kill();  \
  \
   void SetDims(long n, long m);  \
  \
   long NumRows() const { return rep.length(); }  \
   long NumCols() const { return numcols; }  \
  \
   vec_T& operator[](long i) { return rep[i]; }  \
   const vec_T& operator[](long i) const { return rep[i]; }  \
  \
   vec_T& operator()(long i) { return rep[i-1]; }  \
   const vec_T& operator()(long i) const { return rep[i-1]; }  \
  \
   T& operator()(long i, long j) { return rep[i-1][j-1]; }  \
   const T& operator()(long i, long j) const   \
      { return rep[i-1][j-1]; }  \
  \
   friend const vec_vec_T& rep(const mat_T& a)  \
      { return a.rep; }  \
  \
   friend void swap(mat_T& X, mat_T& Y); \
  \
   long position(const vec_T& a) const { return rep.position(a); } \
  mat_T(mat_T& x, INIT_TRANS_TYPE) :  \
    rep(x.rep, INIT_TRANS), numcols(x.numcols) { }  \
};  \
  \
void MakeMatrix(mat_T& x, const vec_vec_T& a);  \



#define ntl_eq_matrix_decl(T,vec_T,vec_vec_T,mat_T) \
long operator==(const mat_T& a, const mat_T& b); \
long operator!=(const mat_T& a, const mat_T& b); \



#define ntl_io_matrix_decl(T,vec_T,vec_vec_T,mat_T) \
istream& operator>>(istream&, mat_T&); \
ostream& operator<<(ostream&, const mat_T&);  \


#define ntl_matrix_impl(T,vec_T,vec_vec_T,mat_T)  \
mat_T::mat_T(const mat_T& a)  \
{  \
   numcols = 0;  \
   SetDims(a.NumRows(), a.NumCols());  \
   rep = a.rep;  \
}  \
  \
mat_T& mat_T::operator=(const mat_T& a)  \
{  \
   SetDims(a.NumRows(), a.NumCols());  \
   rep = a.rep;  \
   return *this;  \
}  \
  \
  \
mat_T::mat_T(INIT_SIZE_TYPE, long n, long m)  \
{  \
   numcols = 0;  \
   SetDims(n, m);  \
}  \
  \
void mat_T::kill()  \
{  \
   numcols = 0;  \
   rep.kill();  \
}  \
  \
void mat_T::SetDims(long n, long m)  \
{  \
   if (n < 0 || m < 0)  \
      Error("SetDims: bad args");  \
  \
   if (m != numcols) {  \
      rep.kill();  \
      numcols = m;  \
   }  \
        \
   long oldmax = rep.MaxLength();  \
   long i;  \
   rep.SetLength(n);  \
  \
   for (i = oldmax; i < n; i++)  \
      rep[i].FixLength(m);  \
}  \
     \
        \
void MakeMatrix(mat_T& x, const vec_vec_T& a)  \
{  \
   long n = a.length();  \
  \
   if (n == 0) {  \
      x.SetDims(0, 0);  \
      return;  \
   }  \
  \
   long m = a[0].length();  \
   long i;  \
  \
   for (i = 1; i < n; i++)  \
      if (a[i].length() != m)  \
         Error("nonrectangular matrix");  \
  \
   x.SetDims(n, m);  \
   for (i = 0; i < n; i++)  \
      x[i] = a[i];  \
}  \
  \
void swap(mat_T& X, mat_T& Y)  \
{  \
   swap(X.numcols, Y.numcols);  \
   swap(X.rep, Y.rep);  \
}  \
  \




   

#define ntl_eq_matrix_impl(T,vec_T,vec_vec_T,mat_T)  \
long operator==(const mat_T& a, const mat_T& b)  \
{  \
   if (a.NumCols() != b.NumCols())  \
      return 0;  \
  \
   if (a.NumRows() != b.NumRows())  \
      return 0;  \
  \
   long n = a.NumRows();  \
   long i;  \
  \
   for (i = 0; i < n; i++)  \
      if (a[i] != b[i])  \
         return 0;  \
  \
   return 1;  \
}  \
  \
  \
long operator!=(const mat_T& a, const mat_T& b)  \
{  \
   return !(a == b);  \
}  \




#define ntl_io_matrix_impl(T,vec_T,vec_vec_T,mat_T)  \
istream& operator>>(istream& s, mat_T& x)  \
{  \
   vec_vec_T buf;  \
   s >> buf;  \
   MakeMatrix(x, buf);  \
   return s;  \
}  \
  \
ostream& operator<<(ostream& s, const mat_T& a)  \
{  \
   long n = a.NumRows();  \
   long i;  \
   s << "[";  \
   for (i = 0; i < n; i++) {  \
      s << a[i]; \
      s << "\n"; \
   }  \
   s << "]";  \
   return s;  \
}  \




#endif
